﻿using System;
using System.Diagnostics;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Windows.Forms;

namespace Mystique
{
	public partial class Mystique:Form
	{
		#region Program data

		public const string Mystique_URL = "http://www.frank-t-clark.com/Professional/Papers/Mystique/Mystique.html";

		Random rnd = new Random ();

		/// <summary> Do we cycle this time
		/// </summary>
		bool cycle = false;

		/// <summary> A Settings shortcut
		/// </summary>
		Properties.Settings Settings = Properties.Settings.Default;

		// Screen coordinates of our form client area
		Rectangle limit;

		/// <summary> A line that has been or will be drawn
		/// </summary>
		struct LINE
		{
			public Point p1; // Screen coordinates
			public Point p2; // Screen coordinates
			public Mirror mirror;
			public Color color;
		};
		LINE[] lines = new LINE[0];
		int current_line = 0;
		Point current_p1;
		Point current_p2;
		Color current_color;

		public Point delta1;
		public Point delta2;

		enum Mirror
		{
			Variable,
			One,
			Two,
			Four,
			Eight,
		};
		Mirror current_mirror;

		enum Wipe
		{
			None,
			Box,
			Octagon,
			Horizontal,
			Vertical
		};
		Wipe current_wipe;
		int current_wipe_index = 0;
		Color old_wipe_color;
		Color current_wipe_color;
		Wipe next_wipe;
		DateTime last_wipe = DateTime.Now;

		/// <summary> Timer for speed control
		/// </summary>
		public Timer speedTimer = new System.Windows.Forms.Timer ();

		#endregion
		#region Program Startup
		/// <summary> The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main (string[] args)
		{
			Application.Run (new Mystique ());
		}
		#endregion
		#region Form code
		public Mystique ()
		{
			InitializeComponent ();
		}
		private void Mystique_Load (object sender, EventArgs e)
		{
			// Adjust to user preference
			if (Settings.WindowPosition.IsEmpty)
				Settings.WindowPosition = Bounds;
			else
				Bounds = Settings.WindowPosition;
			this.Icon = Icon.ExtractAssociatedIcon (Application.ExecutablePath);
			speedTimer.Tick += new System.EventHandler (this.DrawLine);
		}
		private void Mystique_Shown (object sender, EventArgs e)
		{
			// release auto settings to update size/location change
			StartPosition = FormStartPosition.Manual;
			menuViewAuto.Checked = Settings.AutoPattern;
			menuViewCycle.Checked = Settings.ColorCycle;
			ChangeSpeed (Settings.Speed);
			ChangeMirror (Settings.Mirror);
			ChangeWipe (Settings.Wipe);
		}
		private void Mystique_LocationChanged (object sender, EventArgs e)
		{
			// not final settings yet
			if (FormStartPosition.Manual != StartPosition)
				return;
			// don't save minimized
			if (FormWindowState.Minimized == WindowState)
			{
				DrawingToggle (false);
				return;
			}
			// don't save maximized
			if (FormWindowState.Maximized == WindowState)
				return;
			// location changes before and is the same as size changes too often!
			Settings.WindowPosition = Bounds;
			Settings.Save ();
			menuViewInitialize_Click (null, null);
		}
		private void Mystique_Paint (object sender, PaintEventArgs e)
		{
			if (!this.ContainsFocus)
				return;
			menuViewInitialize_Click (null, null);
		}
		private void Mystique_Activated (object sender, EventArgs e)
		{
#if !DEBUG || true
			DrawingStart ();
#endif
		}
		private void Mystique_Deactivate (object sender, EventArgs e)
		{
#if !DEBUG || true
			DrawingToggle (false);
#endif
		}
		private void Mystique_MouseClick (object sender, MouseEventArgs e)
		{
			if (MouseButtons.Left == e.Button)
				DrawingToggle (true);
			else if (MouseButtons.Right == e.Button)
				MaxScreen (true);
		}
		private void Mystique_FormClosing (object sender, FormClosingEventArgs e)
		{
			DrawingToggle (false);
			speedTimer = null;
		}
		#endregion
		#region File Menu Items
		private void menuFile_Popup (object sender, EventArgs e)
		{
			DrawingToggle (false);
		}
		private void menuFileExit_Click (object sender, EventArgs e)
		{
			Application.Exit ();
		}
		#endregion
		#region View Menu Items
		private void menuView_Popup (object sender, EventArgs e)
		{
			DrawingToggle (false);
		}
		private void menuViewInitialize_Click (object sender, EventArgs e)
		{
			DrawingToggle (false);
			Graphics g = CreateGraphics ();
			g.Clear (old_wipe_color = RandomColor (Color.Black, Color.Black));
			limit = this.RectangleToScreen (this.ClientRectangle);
			limit.Width--;
			limit.Height--;
			lines = new LINE[limit.Height];
			current_line = 0;
			current_p1 = RandomPoint (limit);
			current_p2 = RandomPoint (limit);
			current_color = RandomColor (old_wipe_color, old_wipe_color);
			RandomDelta ();
			current_wipe = Wipe.None;
			NewWipe ();
			DrawingStart ();
		}
		private void menuViewAuto_Click (object sender, EventArgs e)
		{
			Settings.AutoPattern = !menuViewAuto.Checked;
			menuViewAuto.Checked = Settings.AutoPattern;
			Settings.Save ();
			DrawingStart ();
		}
		private void menuViewCycle_Click (object sender, EventArgs e)
		{
			Settings.ColorCycle = !menuViewCycle.Checked;
			menuViewCycle.Checked = Settings.ColorCycle;
			Settings.Save ();
			DrawingStart ();
		}
		private void menuViewPause_Click (object sender, EventArgs e)
		{
			menuViewPause.Checked = !menuViewPause.Checked;
			DrawingStart ();
		}
		private void menuViewMax_Click (object sender, EventArgs e)
		{
			MaxScreen (true);
		}
		private void menuViewStop_Click (object sender, EventArgs e)
		{
			DrawingToggle (true);
		}
		#endregion
		#region Speed Menu Items
		private void menuSpeed_Popup (object sender, EventArgs e)
		{
			DrawingToggle (false);
		}
		private void menuSpeed1_Click (object sender, EventArgs e)
		{
			ChangeSpeed (1);
		}
		private void menuSpeed2_Click (object sender, EventArgs e)
		{
			ChangeSpeed (2);
		}
		private void menuSpeed3_Click (object sender, EventArgs e)
		{
			ChangeSpeed (3);
		}
		private void menuSpeed4_Click (object sender, EventArgs e)
		{
			ChangeSpeed (4);
		}
		private void menuSpeed5_Click (object sender, EventArgs e)
		{
			ChangeSpeed (5);
		}
		private void menuSpeed6_Click (object sender, EventArgs e)
		{
			ChangeSpeed (6);
		}
		private void menuSpeed7_Click (object sender, EventArgs e)
		{
			ChangeSpeed (7);
		}
		private void menuSpeed8_Click (object sender, EventArgs e)
		{
			ChangeSpeed (8);
		}
		private void menuSpeed9_Click (object sender, EventArgs e)
		{
			ChangeSpeed (9);
		}
		#endregion
		#region Mirror Menu Items
		private void menuMirror_Popup (object sender, EventArgs e)
		{
			DrawingToggle (false);
		}
		private void menuMirrorVariable_Click (object sender, EventArgs e)
		{
			ChangeMirror ((int) Mirror.Variable);
		}
		private void menuMirror1_Click (object sender, EventArgs e)
		{
			ChangeMirror ((int) Mirror.One);
		}
		private void menuMirror2_Click (object sender, EventArgs e)
		{
			ChangeMirror ((int) Mirror.Two);
		}
		private void menuMirror4_Click (object sender, EventArgs e)
		{
			ChangeMirror ((int) Mirror.Four);
		}
		private void menuMirror8_Click (object sender, EventArgs e)
		{
			ChangeMirror ((int) Mirror.Eight);
		}
		#endregion
		#region Wipe Menu Items
		private void menuWipe_Popup (object sender, EventArgs e)
		{
			DrawingToggle (false);
		}
		private void menuWipeNone_Click (object sender, EventArgs e)
		{
			ChangeWipe ((int) Wipe.None);
		}
		private void menuWipeBox_Click (object sender, EventArgs e)
		{
			ChangeWipe ((int) Wipe.Box);
		}
		private void menuWipeOctagon_Click (object sender, EventArgs e)
		{
			ChangeWipe ((int) Wipe.Octagon);
		}
		private void menuWipeHorizontal_Click (object sender, EventArgs e)
		{
			ChangeWipe ((int) Wipe.Horizontal);
		}
		private void menuWipeVertical_Click (object sender, EventArgs e)
		{
			ChangeWipe ((int) Wipe.Vertical);
		}
		#endregion
		#region Help Menu Items
		private void menuHelp_Popup (object sender, EventArgs e)
		{
			DrawingToggle (false);
		}
		private void menuHelpOnline_Click (object sender, EventArgs e)
		{
			Process process = new Process ();
			process.StartInfo.FileName = Mystique_URL;
			bool result = process.Start ();
		}
		private void menuHelpLocal_Click (object sender, EventArgs e)
		{
			string path = AppDomain.CurrentDomain.BaseDirectory + "\\Mystique.html";
			if (!File.Exists (path))
			{
				string[] lines = { "Open online location", Mystique_URL, "and save a local copy to", path };
				System.IO.File.WriteAllLines (path, lines);
			}
			Process process = new Process ();
			process.StartInfo.FileName = path;
			bool result = process.Start ();
		}
		private void menuHelpAbout_Click (object sender, EventArgs e)
		{
			AboutBox form = new AboutBox ();
			form.Font = new Font (SystemFonts.MessageBoxFont, SystemFonts.MessageBoxFont.Style);
			form.ShowDialog (this);
			DrawingStart ();
		}
		#endregion
		#region speedTimer drawing code
		/// <summary> Drawing state routine
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void DrawLine (object sender, EventArgs e)
		{
			if (0 == lines.Length)
				return;
			DrawWipe ();
			int i = 9 == Settings.Speed ? 8 : 1;
			for (; 0 < i; --i)
			{
				MirrorLine ();	// remove old
				lines[current_line].p1 = current_p1;
				lines[current_line].p2 = current_p2;
				lines[current_line].color = current_color;
				lines[current_line].mirror = current_mirror;
				MirrorLine ();	// draw new
				++current_line;
				// Cycle time?
				if (current_line >= lines.Length)
				{
					current_line = 0;
					cycle = !cycle;
					if (cycle)
					{
						if (menuViewAuto.Checked)
							RandomDelta ();
						if (menuViewCycle.Checked)
							current_color = RandomColor (old_wipe_color, current_wipe_color);
						if (menuMirrorVariable.Checked)
							current_mirror = (Mirror) rnd.Next ((int) Mirror.One, (int) Mirror.Eight);
						if (menuViewPause.Checked)
						{
							DrawingToggle (false);
							return;
						}
					}
				}
				// Next line
				Point p = current_p1;
				p.Offset (delta1);
				if (!limit.Contains (p))
				{
					if (OutRange (p.X, limit.X, limit.X + limit.Width))
						delta1.X = -delta1.X;
					if (OutRange (p.Y, limit.Y, limit.Y + limit.Height))
						delta1.Y = -delta1.Y;
				}
				current_p1.Offset (delta1);
				p = current_p2;
				p.Offset (delta2);
				if (!limit.Contains (p))
				{
					if (OutRange (p.X, limit.X, limit.X + limit.Width))
						delta2.X = -delta2.X;
					if (OutRange (p.Y, limit.Y, limit.Y + limit.Height))
						delta2.Y = -delta2.Y;
				}
				current_p2.Offset (delta2);
			}
		}
		#endregion
		#region Settings
		void ChangeSpeed (int new_speed)
		{
			if (OutRange (new_speed, 1, 9))
				return;
			Settings.Speed = new_speed;
			Settings.Save ();
			if (7 < new_speed)
				speedTimer.Interval = 1;
			else
				speedTimer.Interval = 10 << (8 - new_speed);
			DrawingStart ();
			menuSpeed1.Checked = 1 == new_speed;
			menuSpeed2.Checked = 2 == new_speed;
			menuSpeed3.Checked = 3 == new_speed;
			menuSpeed4.Checked = 4 == new_speed;
			menuSpeed5.Checked = 5 == new_speed;
			menuSpeed6.Checked = 6 == new_speed;
			menuSpeed7.Checked = 7 == new_speed;
			menuSpeed8.Checked = 8 == new_speed;
			menuSpeed9.Checked = 9 == new_speed;
		}
		void ChangeMirror (int new_mirror)
		{
			if (OutRange (new_mirror, (int) Mirror.Variable, (int) Mirror.Eight))
				return;
			Settings.Mirror = new_mirror;
			Settings.Save ();
			if (0 < new_mirror)
				current_mirror = (Mirror) new_mirror;
			this.menuMirrorVariable.Checked = 0 == new_mirror;
			this.menuMirror1.Checked = 1 == new_mirror;
			this.menuMirror2.Checked = 2 == new_mirror;
			this.menuMirror4.Checked = 3 == new_mirror;
			this.menuMirror8.Checked = 4 == new_mirror;
			DrawingStart ();
		}
		void ChangeWipe (int new_wipe)
		{
			if (OutRange (new_wipe, (int) Wipe.None, (int) Wipe.Vertical))
				return;
			Settings.Wipe = new_wipe;
			Settings.Save ();
			next_wipe = (Wipe) new_wipe;
			if (Wipe.None == current_wipe)
				NewWipe ();
			this.menuWipeNone.Checked = 0 == new_wipe;
			this.menuWipeBox.Checked = 1 == new_wipe;
			this.menuWipeOctagon.Checked = 2 == new_wipe;
			this.menuWipeHorizontal.Checked = 3 == new_wipe;
			this.menuWipeVertical.Checked = 4 == new_wipe;
			DrawingStart ();
		}
		#endregion
		#region Helper functions
		/// <summary> Return a random point within a rectangle.
		/// </summary>
		/// <returns></returns>		
		Point RandomPoint (Rectangle rectangle)
		{
			Point point = new Point ();
			point.X = rnd.Next (rectangle.X, rectangle.X + rectangle.Width);
			point.Y = rnd.Next (rectangle.Y, rectangle.Y + rectangle.Height);
			return point;
		}
		/// <summary> Set a useful random delta. Don't allow too little movement.
		/// </summary>
		void RandomDelta ()
		{
			int min = limit.Height / 2;
			for (; ; )
			{
				delta1.X = rnd.Next (-3, 3);
				delta1.Y = rnd.Next (-3, 3);
				delta2.X = rnd.Next (-3, 3);
				delta2.Y = rnd.Next (-3, 3);
				// no movement
				if (delta1.IsEmpty && delta2.IsEmpty)
					continue;
				// little movement and too little separation
				if (delta1.IsEmpty && 0 == delta2.X && min > current_p1.X - current_p2.X)
					continue;
				// little movement and too little separation
				if (delta1.IsEmpty && 0 == delta2.Y && min > current_p1.Y - current_p2.Y)
					continue;
				// little movement and too little separation
				if (delta2.IsEmpty && 0 == delta1.X && min > current_p1.X - current_p2.X)
					continue;
				// little movement and too little separation
				if (delta2.IsEmpty && 0 == delta1.Y && min > current_p1.Y - current_p2.Y)
					continue;
				// same angle and too little separation
				if (delta1.X == delta2.X && min > current_p1.X - current_p2.X)
					continue;
				// same angle and too little separation
				if (delta1.Y == delta2.Y && min > current_p1.Y - current_p2.Y)
					continue;
				break;
			}
		}
		/// <summary> Return a random color for XOR that is not too dark.
		/// Check the excludes such as current background and current_color.
		/// </summary>
		/// <returns></returns>
		Color RandomColor (Color ex1, Color ex2)
		{
			Color color;
			for (; ; )
			{
				int red = rnd.Next (0, 255);
				int green = rnd.Next (0, 255);
				int blue = rnd.Next (0, 255);
				color = Color.FromArgb (red, green, blue);
				if (DarkTest (color, ex1))
					continue;
				if (DarkTest (color, ex2))
					continue;
				break;
			}
			return color;
		}
		/// <summary> Is a color too dark against a background?
		/// </summary>
		bool DarkTest (Color color, Color background)
		{
			int total;
			total = color.R ^ background.R;
			total += color.G ^ background.G;
			total += color.B ^ background.B;
			if (380 > total)
				return true;
			return false;
		}
		/// <summary> Is a value out of the inclusive range
		/// </summary>
		/// <param name="i"></param>
		/// <param name="low"></param>
		/// <param name="high"></param>
		/// <returns></returns>
		bool OutRange (int i, int low, int high)
		{
			if (i > high)
				return true;
			if (i < low)
				return true;
			return false;
		}
		/// <summary> Max Screen Off or toggle
		/// </summary>
		/// <param name="toggle"></param>
		void MaxScreen (bool toggle)
		{
			if (toggle)
				menuViewMax.Checked = !menuViewMax.Checked;
			else
				menuViewMax.Checked = false;
			if (menuViewMax.Checked)
			{
				this.FormBorderStyle = FormBorderStyle.None;
				this.Menu = null;
			}
			else
			{
				this.FormBorderStyle = FormBorderStyle.Sizable;
				this.Menu = mainMenu;
			}
		}
		/// <summary> Speed Off then toggle to start
		/// </summary>
		void DrawingStart ()
		{
			DrawingToggle (false);
			DrawingToggle (true);
		}
		/// <summary> Speed Off or toggle
		/// </summary>
		/// <param name="toggle"></param>
		void DrawingToggle (bool toggle)
		{
			if (null == speedTimer)
				return;
			if (toggle)
				menuViewStop.Checked = !menuViewStop.Checked;
			else
				menuViewStop.Checked = true;
			if (menuViewStop.Checked)
				speedTimer.Stop ();
			else
				speedTimer.Start ();
		}
		/// <summary> Draw the current line with mirrors.
		/// </summary>
		void MirrorLine ()
		{
			// skip when not yet initialized
			if (0 == lines[current_line].color.ToArgb ())
				return;
			Color color = lines[current_line].color;
			Point p1 = lines[current_line].p1;
			Point p2 = lines[current_line].p2;
			int top = p1.Y - limit.Y;
			int left = p1.X - limit.X;
			int bottom = p2.Y - limit.Y;
			int right = p2.X - limit.X;
			p1 = new Point (limit.Left + left, limit.Top + top);
			p2 = new Point (limit.Left + right, limit.Top + bottom);
			ControlPaint.DrawReversibleLine (p1, p2, color);
			if (Mirror.One == lines[current_line].mirror)
				return;
			p1 = new Point (limit.Right - left, limit.Bottom - top);
			p2 = new Point (limit.Right - right, limit.Bottom - bottom);
			ControlPaint.DrawReversibleLine (p1, p2, color);
			if (Mirror.Two == lines[current_line].mirror)
				return;
			p1 = new Point (limit.Left + left, limit.Bottom - top);
			p2 = new Point (limit.Left + right, limit.Bottom - bottom);
			ControlPaint.DrawReversibleLine (p1, p2, color);
			p1 = new Point (limit.Right - left, limit.Top + top);
			p2 = new Point (limit.Right - right, limit.Top + bottom);
			ControlPaint.DrawReversibleLine (p1, p2, color);
			if (Mirror.Four == lines[current_line].mirror)
				return;
			p1 = new Point (limit.Right - left, limit.Bottom - top);
			p2 = new Point (limit.Left + right, limit.Top + bottom);
			ControlPaint.DrawReversibleLine (p1, p2, color);
			p1 = new Point (limit.Left + left, limit.Top + top);
			p2 = new Point (limit.Right - right, limit.Bottom - bottom);
			ControlPaint.DrawReversibleLine (p1, p2, color);
			p1 = new Point (limit.Right - left, limit.Top + top);
			p2 = new Point (limit.Left + right, limit.Bottom - bottom);
			ControlPaint.DrawReversibleLine (p1, p2, color);
			p1 = new Point (limit.Left + left, limit.Bottom - top);
			p2 = new Point (limit.Right - right, limit.Top + bottom);
			ControlPaint.DrawReversibleLine (p1, p2, color);
		}
		/// <summary> Draw the current wipe.
		/// </summary>
		void DrawWipe ()
		{
			switch (current_wipe)
			{
				case Wipe.Box:
					int x1 = limit.X + current_wipe_index;
					int y1 = limit.Y + current_wipe_index;
					int x2 = limit.Right - current_wipe_index;
					int y2 = limit.Bottom - current_wipe_index;
					Point p1 = new Point (x1, y1);
					Point p2 = new Point (x2 + 1, y1);
					ControlPaint.DrawReversibleLine (p1, p2, current_wipe_color);
					if (y1 == y2)
					{
						NewWipe ();
						return;
					}
					p1 = new Point (x1, y2);
					p2 = new Point (x2 + 1, y2);
					ControlPaint.DrawReversibleLine (p1, p2, current_wipe_color);
					++y1;
					if (y2 == y1)
					{
						NewWipe ();
						return;
					}
					p1 = new Point (x1, y1);
					p2 = new Point (x1, y2);
					ControlPaint.DrawReversibleLine (p1, p2, current_wipe_color);
					if (x1 == x2)
					{
						NewWipe ();
						return;
					}
					p1 = new Point (x2, y1);
					p2 = new Point (x2, y2);
					ControlPaint.DrawReversibleLine (p1, p2, current_wipe_color);
					if (x1 + 1 == x2)
					{
						NewWipe ();
						return;
					}
					current_wipe_index++;
					break;
				case Wipe.Octagon:
					break;
				case Wipe.Horizontal:
					break;
				case Wipe.Vertical:
					break;
			}
		}
		/// <summary> Start a new wipe
		/// </summary>
		void NewWipe ()
		{
			//			Console.WriteLine (Settings.Speed+" "+ (DateTime.Now-last_wipe) );
			//			last_wipe = DateTime.Now;
			current_wipe = next_wipe;
			current_wipe_index = 0;
			DarkTest (current_wipe_color, old_wipe_color);
			old_wipe_color = current_wipe_color;
			current_wipe_color = RandomColor (current_color, old_wipe_color);
		}
		#endregion
	}
}